/*
 * Copyright (C) 2008 Codership Oy <info@codership.com>
 *
 * $Id$
 */

#ifndef __gcs_core_test__
#define __gcs_core_test__

#include <check.h>

extern Suite *gcs_core_suite(void);

#endif /* __gu_core_test__ */
